#include "calibration.h"


Calibration::Calibration(int size_x, int size_y, int size_window, 
	       double min_corr, char *device, int preview) : QDialog() {

  setupUi(this);

  connect(btAdd, SIGNAL(clicked()), this, SLOT(takePicture()));
  connect(btRemove, SIGNAL(clicked()), this, SLOT(removePicture()));
  connect(btColours, SIGNAL(clicked()), this, SLOT(takeColours()));
  connect(btClick, SIGNAL(clicked()), this, SLOT(clicCorners()));
  connect(btCorners, SIGNAL(clicked()), this, SLOT(findCorners()));
  connect(btCalib, SIGNAL(clicked()), this, SLOT(calibrate()));
  connect(listWidget, SIGNAL(currentRowChanged(int)),
	  this, SLOT(changeImage(int)));

  this->size_x = size_x;
  this->size_y = size_y;
  this->size_window = size_window;
  this->min_corr = min_corr;
  this->device = string(device);
  this->preview = preview;

  if(preview) {
    timer = new QTimer( this );
    connect( timer, SIGNAL(timeout()), this, SLOT(refresh_image()) );
    timer->start( 1000 );
  }

  colour = CORNER;

}

void Calibration::refresh_image() {

  try {

    Picture p(device);

    video_image->setPixmap(p.getPixmap().scaledToWidth(120));

  } catch (mimasexception &e) {

    timer->stop();

    QMessageBox::critical(0, QString("Calibration"),
			  QString("Couldn't read the video device."));
    
  }
}


void Calibration::takePicture() {
  
  lbNotes->setText("Click on the four extreme corners or remove the image.");

  if(preview)
    timer->stop();

  static int i = 0;

  try {

    // Get an image from the camera and add the image to the list
    list += Picture(device);

    // convert indice to string
    std::ostringstream oss;
    oss << ++i;
    string s = "Picture " + oss.str();

    listWidget->addItem(QString(s.c_str()));
    listWidget->setCurrentRow(list.size() - 1);

  } catch(mimasexception &e) {

    QMessageBox::critical(0, QString("Calibration"),
			  QString("Couldn't read the video device."));

  }

}


void Calibration::removePicture() {

  lbNotes->setText("Take another image.");

  int current = listWidget->currentRow();

  if(current != -1) {
    list.removeAt(current);
    listWidget->takeItem(current);
  }

  if(list.size() > 0)
    listWidget->setCurrentRow(list.size() - 1);
  else
    lbImage->setDisplayPresent(false);

  if(preview)
    timer->start();

}


void Calibration::clicCorners() {

  lbImage->resetCorners();

}


void Calibration::calibrate() {

  if(preview)
    timer->stop();

  std::vector< Chessboard > chess;

  int n = list.size();

  if(n < 4)
    QMessageBox::critical(0, QString("Calibration"),
			  QString("4 pictures minimum to calibrate the camera."));
  else {

    lbNotes->setText("Calibration started...");

    for(int i = 0; i < n; ++i) {

      Chessboard c;
      Picture p = list.at(i);

      c.image.push_back(pixel_f(p.corns[0].x, p.corns[0].y));
      c.image.push_back(pixel_f(p.corns[1].x, p.corns[1].y));
      c.image.push_back(pixel_f(p.corns[2].x, p.corns[2].y));
      c.image.push_back(pixel_f(p.corns[3].x, p.corns[3].y));

      c.object.push_back(point_f(0, 0));
      c.object.push_back(point_f(2*size_x-1, 0));
      c.object.push_back(point_f(2*size_x-1, 2*size_y-1));
      c.object.push_back(point_f(0, 2*size_y-1));

      chess.push_back(c);

    }

    try {

      Camera cam;

      cam.calibrate(chess);

      Camera::Matrix K = cam.getIntrinsicsInit();

      std::cout << std::endl 
		<< "Intrinsic parameters : (first approximation)" << std::endl;

      for (unsigned i = 0; i < K.size1 (); ++ i) {
	for (unsigned j = 0; j < K.size2 (); ++ j)
	  std::cout << setw(10) << left << K(i, j) << "  ";
	std::cout << std::endl;
      }
      std::cout << std::endl;

      if(cam.isOptimized()) {

	K = cam.getIntrinsics();

	std::cout << std::endl 
		  << "Intrinsic parameters :" << std::endl;
	for (unsigned i = 0; i < K.size1 (); ++ i) {
	  for (unsigned j = 0; j < K.size2 (); ++ j)
	    std::cout << setw(10) << left << K(i, j) << "  ";
	  std::cout << std::endl;
	}
	std::cout << std::endl;

	std::cout << "Distortion :" << std::endl;
	std::cout << "k1 = " << cam.getDistortion()(0) << std::endl
		  << "k2 = " << cam.getDistortion()(1) << std::endl;

	lbNotes->setText("Calibration finished.");
  
      } else {

	std::cerr << "Couldn't calculate the maximum likelihood optimisation" << std::endl;

      }	      

    } catch (mimasexception &e) {

      QMessageBox::critical(0, QString("Calibration"),
			    QString("Couldn't compute the calibration. \n"
				    "Try again with new images."));

    }

  }

}


void Calibration::changeImage(int current) {

  lbNotes->setText("");

  lbImage->setImage(&list[current]);

}


void Calibration::findCorners() {

  lbNotes->setText("Click on the four extreme corners.");

  try {

    lbImage->findCorners(size_x, size_y, size_window, min_corr);

    if(preview)
      timer->start();

  } catch (mimasexception &e) {

    //nothing to do

  }

}


void Calibration::takeColours() {

  if(list.size() == 0) {

    QMessageBox::critical(0, QString("Calibration"),
			  QString("Take a picture first."));
    return;

  } else if (colour == CORNER) {

    colour = BLACK;

    lbNotes->setText("Click on a black square of the chessboard.");
  
    lbImage->click(colour);

  } else if (colour == BLACK) {

    colour = WHITE;

    lbNotes->setText("Click on a white area of the chessboard.");
  
    lbImage->click(colour);

  } else {

    colour = CORNER;

    lbNotes->setText("OK");
  
    lbImage->click(colour);

  }

}
