// Sample application of fourier library functions
// Bala Amavasai (bpa@amavasai.org)
// Thu Oct 18 12:38:26 BST 2001
// Jan Wedekind (jan@wedesoft.de)
// Mon Aug 23 14:37:05 UTC 2004

#include <algorithm>
#include <cmath>
#include <complex>
#include <iostream>
#include <fstream>
#include <valarray>
#include "fourier.h"
#include "image.h"
#include "image_fileinput.h"
#include "image_mesaoutput.h"
#include "multi_array_op.h"

using namespace boost;
using namespace std;
using namespace mimas;

template< typename T >
struct _todouble: public unary_function< T, double >
{
  double operator()( const T &x ) const
  { return static_cast< double >( x ); }
};

template< typename T >
struct _tocomplex: public unary_function< T, complex< double > >
{
  complex< double > operator()( const double &x ) const
  { return complex< double >( x ); }
};

template< typename T >
struct _toint: public unary_function< T, int >
{
  int operator()( const T &x ) const { return static_cast< int >( x ); }
};

template< typename T >
struct _norm: public unary_function< T, double >
{
  double operator()( const T &x ) const { return norm( x ); }
};

template< typename T >
struct _creal: public unary_function< T, double >
{
  double operator()( const T & x ) const { return x.real(); }
};

template< typename T >
struct _log: public unary_function< T, double >
{
  double operator()( const T &x ) const { return log( x ); }
};

ARRAY_OP_EX( toint, _toint )
ARRAY_OP_EX( todouble, _todouble )
ARRAY_OP_EX( tocomplex, _tocomplex )
ARRAY_OP_EX( norm, _norm )
ARRAY_OP_EX( creal, _creal )
ARRAY_OP_EX( log, _log )

template< typename T >
image< T > multi_array_to_image( const multi_array< T, 2 > &x )
{
  image< T > retVal; retVal.init( x.shape()[ 1 ], x.shape()[ 0 ] );
  retVal.rawData() = x;
  return retVal;
}

int main(int argc, char **argv)
{
  int retVal = 0;
  try {
    x11_display display;
    image< int > image;
    image_mesaoutput< unsigned char >
      window1( &display ),
      window2( &display ),
      window3( &display ),
      window4( &display );

    ifstream file( "../images/stills/jellybeans-256x256.pgm", ios::binary );
    file >> image;

    {
      // Real fourier transform.
      multi_array< complex< double >,2 > imageFFT
	( rfft( todouble( image.rawData() ) ) );
      multi_array< double, 2 > imageLogFFT( log( norm( imageFFT ) + 1.0 ) );

      // Remove peak caused by constant component of input image.
      imageLogFFT[ 0 ][ 0 ] = 0.0;

      // Scale to interval [0,1]
      // ... *= ( 1.0 / ... ) would be faster.
      {
	double max =
	  *max_element< double * >( imageLogFFT.data(),
				    imageLogFFT.data() +
				    imageLogFFT.num_elements() );
	imageLogFFT /= max;
      };

      // Save
      window1 << multi_array_to_image( toint( 255.0 * imageLogFFT ) );

      // Inverse fourier transform.
      window2 << multi_array_to_image( toint( invrfft( imageFFT ) ) );
    };
    {
      // Fourier transform.
      multi_array< complex< double >,2 > imageFFT
	( fft( tocomplex( todouble( image.rawData() ) ) ) );
      multi_array< double, 2 > imageLogFFT( log( norm( imageFFT ) + 1.0 ) );

      // Remove peak caused by constant component of input image.
      imageLogFFT[ 0 ][ 0 ] = 0.0;

      // Scale to interval [0,1]
      // ... *= ( 1.0 / ... ) would be faster.
      {
	double max =
	  *max_element( imageLogFFT.data(),
			imageLogFFT.data() + imageLogFFT.num_elements() );
	imageLogFFT /= max;
      };

      // Save
      window3 << multi_array_to_image( toint( 255.0 * imageLogFFT ) );

      // Inverse fourier transform.
      window4 << multi_array_to_image( toint( creal( invfft( imageFFT ) ) ) )
	      << mimas::pause();
    };
  } catch ( exception &e ) {
    cerr << e.what() << endl;
    retVal = 1;
  };
  return retVal;
}
