#include <fstream>
#include <iostream>
#include <iomanip>
#include <sstream>
#include "image.h"
#include "image_fileinput.h"
#include "image_fileoutput.h"
#include "cartesian.h"
#include "correlation_tracking_tool.h"

using namespace mimas;

template <class Sequence>
struct SSD: public std::binary_function< Sequence, Sequence, void >
{
  double operator()( const Sequence &a, const Sequence & b ) {
    double ssd = 0.0;
    assert( a.size() == b.size() );
    for ( typename Sequence::const_iterator i = a.begin(), j = b.begin();
          i != a.end(); i++, j++ )
      ssd -= fabs( (double)( *j - *i ) );
    return ssd  / a.size();
  };
};

using namespace std;

string fileName( const char *dir, const char *imageBaseName, int number )
{
  ostringstream str;
  str << dir << '/' << imageBaseName << '_' << setw( 3 ) << setfill( '0' )
      <<  number << ".pgm";
  return str.str();
}


int main(int argc, char **argv)
{
 int retVal = 0;
 try {
	image<int> imagetemplate;
	string dir, in_images_basename, out_images_basename;
	int
     frames = 0,
     update_rate,
     no_roi = 0,
     no_points = 0,
     xrange, yrange, frame_gray_level;

   MMERROR( argc == 2, mimasexception, ,
            "Syntax : " << argv[0] << " config-file" );

	ifstream in(argv[1]);

	in >> dir >> in_images_basename >> out_images_basename >> frames
      >> update_rate >> no_roi >> no_points;

   ifstream templatefile( fileName( dir.c_str(),
                                    in_images_basename.c_str(), 0 ).
                                    c_str(), ios::binary );
   templatefile >> imagetemplate;

	int x[no_roi][no_points], y[no_roi][no_points];
	correlation_tracking_tool< int, SSD< vector< double > > >
     iTemplate[no_roi];
	Cartesian<int> start[no_roi], range[no_roi];
	
	
	for (int i = 0; i < no_roi; i++)
	{
		iTemplate[i].clearPolyCoords();
	
		for (int j = 0; j < no_points; j++)
		{
         in >> x[i][j] >> y[i][j],
			iTemplate[i].pushPolyCoord(x[i][j],y[i][j]);
		}

		iTemplate[i].createTemplate(imagetemplate);

		start[i].x = x[i][0]; start[i].y = y[i][0];
	}

	
	in >> xrange >> yrange >> frame_gray_level;
   MMERROR( in, mimasexception, ,
            "Error reading from file \"" << argv[1] << "\"." );
	in.close();

	for (int i = 0; i < no_roi; i++)
	{
		range[i].x = xrange; range[i].y = yrange;
	}

   SSD< vector< double > > ssd;

	for (int s = 0; s < frames; s++)
	{
		image<int> sequencetrack;
		Cartesian<int> final[no_roi];
		ifstream sequencefile( fileName( dir.c_str(),
					  in_images_basename.c_str(), s ).
                                          c_str(), ios::binary );
		sequencefile >> sequencetrack;

		for (int i = 0; i < no_roi; i++)
		{
        final[i] = iTemplate[i].trackCorrelation(sequencetrack,
                                                 start[i], range[i],
                                                 ssd );
			cout << "Returned tracking correlation value "
              << iTemplate[i].getTrackingCorr() << endl;
		}

		/* debugging info */
		cout << fileName( dir.c_str(), in_images_basename.c_str(), s ).c_str();
		for (int i = 0; i < no_roi; i++)
			cout << "     " << final[i];
		cout << endl;

		if (update_rate > 0 && s % update_rate == 0)
		{
			for (int i = 0; i < no_roi; i++)
			{
				iTemplate[i].clearPolyCoords();
				for (int j = 0; j < no_points; j++)
					iTemplate[i].pushPolyCoord(final[i].x + x[i][j] - x[i][0], final[i].y + y[i][j] - y[i][0]);
				iTemplate[i].createTemplate(sequencetrack);
			}
		}

		for (int i = 0; i < no_roi; i++)
			iTemplate[i].showPolygonOnImage(sequencetrack, frame_gray_level);

		ofstream outfile( fileName( dir.c_str(), out_images_basename.c_str(), s ).
                                    c_str(), ios::binary );
		outfile << setformat( "PGM" ) << sequencetrack;

		for (int i = 0; i < no_roi; i++)
		{
			start[i].x = final[i].x; start[i].y = final[i].y;
		}
	}
  } catch ( exception &e ) {
    cerr << e.what() << endl;
    retVal = 1;
  };
  return retVal;
}
