// You can only define one of these:
// #define TESTMESAOFFSCREEN
// #define TESTEDGEDETECT

#include <cmath>
#include <fstream>
#include <functional>
#include <GL/glut.h>
#ifdef TESTMESAOFFSCREEN
#include "mesa_offscreen.h"
#include "timer.h"
#endif
#ifdef TESTEDGEDETECT
#include "filter.h"
#include "image_op.h"
#endif
#include "image_xineinput.h"
#include "image_mesaoutput.h"

using namespace boost;
using namespace std;
using namespace mimas;

#ifdef TESTEDGEDETECT
typedef unsigned char pixel_t;
#else
typedef rgba< unsigned char > pixel_t;
#endif


#ifdef TESTEDGEDETECT
template< typename T >
struct _tobw: public std::unary_function< T, unsigned char > {
  unsigned char operator()( const T &x ) const
  { return x ? 255 : 0; }
};

template< typename T >
struct fast_sqr: public std::unary_function< T, int >
{
  fast_sqr(void) {
    for ( short int i=0; i<(signed)table.size(); i++ )
      table[i] = (int)i * i;
  }
  int operator()( const T &x ) const {
    assert( x > -(signed)table.size() && x < (signed)table.size() );
    return table[ x < 0 ? -x : x ];
  }
  boost::array< int, 512 > table;
};

template< typename T >
struct _fastsqr: public std::unary_function< T, int >
{
  int operator()( const T &x ) const { return f( x ); }
  fast_sqr< T > f;
};

ARRAY_OP_EX( fastsqr, _fastsqr )
IMAGE_OP_EX( fastsqr, _fastsqr )
ARRAY_OP_EX( tobw   , _tobw    )
IMAGE_OP_EX( tobw   , _tobw    )

image< short int > gradX( const image< unsigned char > &m )
{
  image< short int > retVal;
  filter< short int >::gradX
    ( image< short int >( m ), retVal );
  return retVal;
}

image< short int > gradY( const image< unsigned char > &m )
{
  image< short int > retVal;
  filter< short int >::gradY
    ( image< short int >( m ), retVal );
  return retVal;
}

void dxvec( short int gx, short int gy, short int w,
            signed char &dx, short int &dy )
{
  short int
    agx = gx < 0 ? -gx : gx,
    agy = gy < 0 ? -gy : gy;
  if ( agy < agx ) {
    if ( ( agy << 1 ) < agx )
      dy = 0;
    else
      dy = gy < 0 ? -w : w;
    dx = gx < 0 ? -1 : 1;
  } else {
    if ( ( agx << 1 ) < agy )
      dx = 0;
    else
      dx = gx < 0 ? -1 : 1;
    dy = gy < 0 ? -w : w;
  };
}

image< bool > edgeDetect( const image< unsigned char > &m,
			  short int threshold )
{
  int thresholdSqr = (int)threshold * threshold;

  image< short int >
    gradientX( gradX( m ) ),
    gradientY( gradY( m ) );

  image< int >
    gradientSqr( fastsqr( gradientX ) + fastsqr( gradientY ) );

  image< bool > retVal; retVal.init( m.getWidth(), m.getHeight() );

  short int w = (short)m.getWidth();
  const int *pgSqr = gradientSqr.rawData().data() + w + 1;
  const short int
    *pgX = gradientX.rawData().data() + w + 1,
    *pgY = gradientY.rawData().data() + w + 1;
  bool *pr = retVal.rawData().data() + w + 1;

  for ( int y=1; y<m.getHeight()-1; y++ ) {
    for ( int x=1; x<m.getWidth()-1; x++ ) {

      signed char dx;
      short int dy;

      dxvec( *pgX, *pgY, m.getWidth(), dx, dy );

      int g = *pgSqr;
      if ( g >= thresholdSqr &&
           g >= pgSqr[ - ( dy + dx ) ] &&
           g >= pgSqr[ + ( dy + dx ) ] )
        *pr = true;

      pgSqr++;
      pgX++;
      pgY++;
      pr++;
    };
    pgSqr += 2;
    pgX += 2;
    pgY += 2;
    pr += 2;
  };

  return retVal;
}
#endif


int main( int argc, char *argv[] )
{
  int retVal = 0;

  try {

    // Note, that this is a very inefficient way to do it, if you just want
    // to display a movie. If you just want to implement a video player, you
    // should allow libxine do write directly to X11 memory!
    // See http://xinehq.de/index.php/hackersguide#AEN80 for more information.

    // Initialise GLUT.
    glutInit( &argc, argv );

    MMERROR( argc == 2, mimasexception, ,
             "Syntax: " << argv[0] << " <media resource location>" );

    // Open X11 display.
    x11_display display;

    const int delay = 100;

    image_mesaoutput< pixel_t > window( &display );
    image_xineinput< pixel_t > input( argv[1], "auto", "auto" );


#ifdef TESTMESAOFFSCREEN
    timer t;
#endif

    // Allocate image.
    image< pixel_t > img;

    // Loop for images.
    do {
      
      try {
        input >> img;

#ifdef TESTMESAOFFSCREEN
        mesa_offscreen< rgba< unsigned char > > m( &display, img, false );
        m.ortho2D();
        const rgba< unsigned char > red( 255, 0, 0 );
        m.setColor( red );

        glPushMatrix();
        double animate = cos( fmod( t.elapsed() * M_PI / 2.0, 2.0 * M_PI ) );
        glTranslated( 10, 10 + animate * 6, 0 );
        glScaled(  0.1,
		  -0.1 * animate,
		   1.0 );
        m.text( "Mimas" );
        glPopMatrix();
        m.finish();
#endif

#ifdef TESTEDGEDETECT
        window << tobw( edgeDetect( img, 15 ) ) << mimas::delay( delay );
#else
        window << img << mimas::delay( delay );
#endif
      } catch ( exception &e ) {
        cerr << "Warning: " << e.what() << endl;
        window << mimas::wait( delay );
      };
      
    } while ( input );

  } catch ( exception &e ) {
    cerr << "Error: " << e.what() << endl;
    retVal = 1;
  };
  
  return retVal;
}
