#ifndef CORRCOEF_H
#define CORRCOEF_H

#include <cassert>
#include <functional>
#include <cmath>

namespace mimas {
/** Compute correlation-value of set of value-tupples.
    This function assigns a real number (the correlation value) to
    two sets of real numbers.
    \f$\{(a_1,b_1),(a_2,b_2),\ldots\}\Rightarrow Corr(A,B)\f$.
    \f$corr(A,B)=\frac{Cov(A,B)}{\sqrt{Cov(A,A)*Cov(B,B)}}\f$
    where \f$Cov(A,B)=E((A-E(A))*(B-E(B)))\f$
    
    The template-parameters \c Arg1 and \c Arg2 need to implement the
    concept of
    <A HREF="http://www.sgi.com/tech/stl/Sequence.html">Sequence</A>.

    The syntax is intended to look like Matlab's \c corrcoef function.

    @todo The interface should be Matlab-like, but there also should be a
    interface for faster access, which doesn't require intermediate
    container-objects.
    @author Jan Wedekind (jan@wedesoft.de)
    @date Wed May 19 12:22:00 UTC 2004 */
template <class Arg1, class Arg2 = Arg1 >
struct corrcoef: public std::binary_function< Arg1, Arg2, double >
{
  double operator()( const Arg1 &a,
                     const Arg2 &b ) {
    double sa = 0.0, sb = 0.0, saa = 0.0, sbb = 0.0, sab = 0.0;
    assert( a.size() == b.size() );
    typename Arg1::const_iterator i;
    typename Arg2::const_iterator j;
    for ( i = a.begin(), j = b.begin(); i != a.end(); i++, j++ ) {
      sa += *i;
      sb += *j;
      saa += *i * *i;
      sbb += *j * *j;
      sab += *i * *j;
    };
    int n = a.size();
    return ( n * sab - sa*sb ) /
      sqrt( ( n * saa - sa * sa ) *
            ( n * sbb - sb * sb ) );
  }
};

}
#endif
