#ifndef GNUPLOT_H
#define GNUPLOT_H

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <cstdarg>
#include <float.h>
#include <iomanip>
#include <sstream>
#include <string>
#include <vector>
#include "image.h"
#include "rgba.h"

namespace mimas {

class gnuplot: public object
{
public:
  /// Constructor.
  gnuplot( image< rgba< unsigned char > > &_img,
           int width, int height, int _timeout = 3 ) throw (mimasexception);
  /** Destructor.
      Will call \c finish(), if it was not called already. */
  virtual ~gnuplot(void);
  /// Render plot to image.
  void finish(void) throw (mimasexception);
  /// Append something to the gnuplot script.
  template< typename T >
  std::ostream &operator<<( const T &t ) { program << t; return program; }
  /// Append io-manipulator to the gnuplot script.
  std::ostream &operator<<( std::ostream& (*__pf)( std::ostream&) )
    { (*__pf)( program ); return program; }
  /// Plot vector - like in Octave
  template< typename T >
  void plot( const boost::numeric::ublas::vector< T > &vec );
  /// Plot matrix - like in Octave
  template< typename T >
  void plot( const boost::numeric::ublas::matrix< T > &mat );
  /** Plotting of xcol against ycol.
      Columns are indexed from 0. style is either "line" or "point". type
      is an integer which specifies the type of line or point (refer to
      gnuplot package). size is the point or line size. Use setMultiPlot
      to overlay plots. */
  template< typename T >
  void plot( const boost::numeric::ublas::matrix< T > &mat,
             int xcol, int ycol,
             const std::string &style = "points",
             int type=1, int size=1 ) throw (mimasexception);
  /// plot an image<int> profile
  template< typename T >
  void plotImageProfile( const image< T > &im );
  /// set or unset multiplot
  void setMultiPlot(bool multiplot);
  /// Set line titles for use with plotxy.
  void setLineTitle(const char *title, ...);
  /// Unset line titles.
  void unsetLineTitle(void);
  /** plotting of x and single or multple y column pairs.
      List of y columns must be terminated by NULL or it will lead to
      indiscriminate behaviour. Columns are indexed from 0. */
  template< typename T >
  void plotxy( const boost::numeric::ublas::matrix< T > &mat,
               int xcol, int ycol, ... ) throw (mimasexception);
protected:
  ///
  static void alarmHandler( int signal );
  /// The plot will be rendered to this image.
  image< rgba< unsigned char > > *img;
  /// Kill 'gnuplot' after so many seconds.
  int timeout;
  /// Boolean indicating, wether finish was called already.
  bool finished;
  /// Process-ID of 'gnuplot'.
  static pid_t pid;
  /// Pipes to communicate with 'gnuplot'.
  int pipes[2];
  /// Stream to temporarily store gnuplot script.
  std::ostringstream program;
  /// List of titles used by plotxy.
  std::vector< std::string > lineTitles;
};

}

#include "gnuplot.tcc"

#endif
