#ifndef MMVL_MIMAS_HYPOTHESIS_FILTER
#define MMVL_MIMAS_HYPOTHESIS_FILTER

#include <vector>
#include <cassert>
#include <cmath> 
#include <iostream>
#include <boost/shared_ptr.hpp>
#include <image.h>

#include "condensation.h"
#include "no_compensation.h"

using std::vector;
using std::cout;
using std::cerr;
using std::endl;
using boost::shared_ptr;

namespace mimas{
  /**@ingroup pf
     @namespace mimas::hf
     @brief Classes related to the hypothesis filter; a tracking method better known as particle filter. 
  */
  namespace hf{
    typedef boost::numeric::ublas::vector<float> Point;


    /**\defgroup pf Particle filter  
       @brief Look at the examples in the examples/particle_filter/* repertories to see how to use 
       particle filter with mimas.
       You will be quickly able to implement basic solution for your tracking problems.
       You may then look at the hypothesis_filter class 
    */

    
    /**@brief Contains the data shared by the different policies of hypothesis filter.

       We used this additionnal class as it allows policies to refer shared data during the
       construction of the hypothesis filter class. These data could not have been put as direct 
       members of hypothesis filter as the variables would have spawn after the whole inheritence
       process that already need these data. For instance, hypothesis filter inherits of image_loader
       that needs a reference to the image variable that will contain the current image data.
       
       This can be seen as a generic trick that allows policies to share data. I don't know if this
       technique is documented in the literature so "si le coeur vous en dit ..."
    
       Said differently: we store in a policy the data that will be shared/used by different policies.
       This data can then be used by the subsequent inherited policies during the
       construction of the host class.
    
       @author Manuel Boissenin
    */
    template<typename Hypothesis_T, typename Pixel_T>
      class hypothesis_filter_data{
      public:
      image<Pixel_T> image;
      std::vector<Hypothesis_T > hypotheses;    
    };

    
    /**
     *\ingroup pf
     *@brief The host class, center of the hypothesis filter architecture. 
     *
     *It hosts different policy classes;
     *if you want to do modifications
     *you should start to examine this class and its policies.
     *
     *If you just want to use this class browse through the provided examples.
     *It might already provide you with an all ready solution.
     *
     *Particle filter is a very versatile method, you can use it to do tracking using active contours, template matching in 2-D, 3-D...
     *To take this into account we developed a generic implementation that shall adapt to a broad range of situations 
     *and that should save time to the programmer by providing the basic bricks common to different implementations.
     *We thus used the policy idiom which is arguably faster than usual polymorphism since links are established at compile time. It
     *is yet more flexible as it is possible to add unforcast functionalities to our tracking object without deep modifications. 
     *As a result the code might be more complex for the library programmer however the user get easier life as you may convince  
     *yourself by looking at the examples.
     *
     *If you want to add building block to the library you may continue to read.
     *
     *Following, the "raison d'tre" of the different template parametres, what they implement and wether they are necessary for the hypothesis_filter class to work:
     *- image_loader policy, to load the sequence of images from where the object of interest is tracked. next()
     *- observer policy, to specify how the object presence is measured for a given location.To evaluate the weight of a particule. An observer policy is used. 
     *The validity of the hypothesis (its weight) is measured by the observer.
     *The observer has to give positive weight.
     *This function can be seen as an interface between the filter class and the observer.
     *Hence it is adapted to check here if contract between the two objects is respected.
     *the user is left with the displaying of the image,
     *the image member can be used for this effect,
     *a draw function that might be provided by the
     *observer policiy can be used to display result on the image.
     * The Observer is what will give you the measures on your image. The choice of your observer
     *is fundamental for the result of the filtering.
     *
     *The number of hypothesis you have to take depends a lot on
     *the choice of your observer and the system you are observing.
     *To give you an idea, for the
     *first application of this class I was using 1000.
     *
     *As a rule of thumbs, the more hypothesis you have, the better your filtering and the slowler the computation.
     *@sa correlation_observer2D 
     *- analyse_result policy, to determine how the collected information is filtered to infer the location of the object. To this effect,
     *it provides analyse() which post-process the information extracted by the particle filter (without necessarily modifying it). If the function is not
     *implemented or has a different name it won't affect the hypothesis_filter class.
     *@sa 
     * - analyse_result_mean
     * - analyse_result_blob
     *- hypothesis template to implement the hypothesis concept and how hypotheses are drifted @sa hypothesis2D
     *- picking policy,  to implement the way of selecting the hypothesis that will be used to
     *derive the new hypothesis. @sa condensation
     *- compensation policy (cf no_compensation)
     *- pixel template to specify the type of data we are working with (color/grey scale images) (ie rgba<unsigned char>)
     *
     *@sa 
     *- For a theoretical background of particule filtering you can read:
     *CONDENSATION- conditionnal density propagation for visual tracking.
     *by Michael Isard and Andrew Blake.
     *You can go to the condensation algorithm homepage to have an idea of 
     *what you can do with hypothesis filtering:
     *http://www.dai.ed.ac.uk/CVonline/LOCAL_COPIES/ISARD1/condensation.html   
     *- Selvan, A.N., Boissenin, M., Amavasai, B.P. and Travis J.R., 
     *Tracking Translucent Objects in Cluttered Scenes
     *IEEE Conference on Cybernetics Intelligence
     *Challenges and Advances, pp. 110-118, Reading, UK, 17 Sept, 2003.
     *- The policy idiom on the Internet. It explains why so many templates are used. 
     *
     *@author Manuel Boissenin 
     *@date June 2003
     *****************************************/
    
    template <
      template <typename> class image_loader_P,
      template <typename, typename> class Observer_P,      //<provides double observe(Hypothesis_T h)
      template<typename, typename> class Analyse_result_T, //<provides analyse()
      typename Hypothesis_T,
      template <typename> class Picking_P = condensation,  
      template <typename> class Compensate_P = no_compensation,
      typename Pixel_T = unsigned char
      >
      class hypothesis_filter:
      public hypothesis_filter_data<Hypothesis_T, Pixel_T>, 
      public image_loader_P<Pixel_T>,
      public Observer_P<Hypothesis_T, Pixel_T>,
      public Analyse_result_T<Hypothesis_T, Pixel_T>,
      public Picking_P<Hypothesis_T>,
      public Compensate_P<Hypothesis_T>
      { 
	public:
	///misc
	/** gives the maximum weight of the particules, can serve to normalise measures for instance. Complexity: O(nb_hypothesis)*/ 
	double max_weight(void);

	/**This method is implemented by the compiler only if it is called.
	However policies won't necessarily provide the necessary method (here next() 
	or analyse()). The compiler will tell you if it is the case.
	You may use this method as a reference how to use the hypothesis filter class
	to track.*/
	void track(void)
	{
	  while(true){
	    try{image_loader_P<Pixel_T>::next();}
	    catch (mimasexception){return;}
	    
	    Hypothesis_T estimatedLocation;
	    filter();
	    estimatedLocation = Analyse_result_T<Hypothesis_T, Pixel_T>::analyse();
	  
	    std::cerr<<"estim: "
		     <<estimatedLocation.x
		     <<estimatedLocation.y
		     <<std::endl;

	    draw(estimatedLocation);                     //from Observer
	    save_location(estimatedLocation);            //from Compensate policy 
	  }
	}


	/**To re-initialize the filter, very usefull to regroup around one hypothesis the different
	   hypothesis when they are wide-spreaded and the measurment is "good enough".
	   May reduce the robustness of your tracker if used.
	*/
	void reset(Hypothesis_T &init);
	
	/**When all has been set call this method to measure hypothesis validity.
	 The behaviour of this method is affected by the observer policy, the compensate policy,
	the picking policy, and the drift method of the hypothesis.
	*/
	void filter(void);
	
	/**The constructor, you right it looks awful but it does a hell of a lot for you*/
	hypothesis_filter(Hypothesis_T &initial,
			  int n,
			  typename image_loader_P<Pixel_T>::set il_param,
			  typename Observer_P<Hypothesis_T, Pixel_T>::set obs_param, 
			  typename Analyse_result_T<Hypothesis_T, Pixel_T>::set analyse_param)
	:image_loader_P<Pixel_T>(hypothesis_filter_data<Hypothesis_T, Pixel_T>::image,il_param),
	Observer_P<Hypothesis_T, Pixel_T>(hypothesis_filter_data<Hypothesis_T, Pixel_T>::image, obs_param),
	Analyse_result_T<Hypothesis_T, Pixel_T>(hypothesis_filter_data<Hypothesis_T, Pixel_T>::hypotheses, hypothesis_filter_data<Hypothesis_T, Pixel_T>::image, analyse_param),
	Picking_P<Hypothesis_T>(hypothesis_filter_data<Hypothesis_T, Pixel_T>::hypotheses),
	nb_hypothesis(n) 
	{
	  for(int i = 0; i < n; ++i)
	    hypothesis_filter_data<Hypothesis_T, Pixel_T>::hypotheses.push_back(initial.drift());
	};
	
	//if we don't have parameter to pass to the observer or the analyse policies we can simplify the instanciation of an hypothesis_filter class
	hypothesis_filter(Hypothesis_T &initial,
			  int n,
			  typename image_loader_P<Pixel_T>::set il_param)
	:image_loader_P<Pixel_T>(hypothesis_filter_data<Hypothesis_T, Pixel_T>::image,il_param),
	Observer_P<Hypothesis_T, Pixel_T>(hypothesis_filter_data<Hypothesis_T, Pixel_T>::image, typename Observer_P<Hypothesis_T, Pixel_T>::set()),
	Analyse_result_T<Hypothesis_T, Pixel_T>(hypothesis_filter_data<Hypothesis_T, Pixel_T>::hypotheses, hypothesis_filter_data<Hypothesis_T, Pixel_T>::image, typename Analyse_result_T<Hypothesis_T, Pixel_T>::set()),
	Picking_P<Hypothesis_T>(hypothesis_filter_data<Hypothesis_T, Pixel_T>::hypotheses),
	nb_hypothesis(n) 
	{
	  for(int i = 0; i < n; ++i)
	    hypothesis_filter_data<Hypothesis_T, Pixel_T>::hypotheses.push_back(initial.drift());
	};
	
	~hypothesis_filter(){}
	protected:
	///for plublic access look at the vector size
	///record the number of hypothesis used for re-initialisation 
	int nb_hypothesis;	

	private:  

	vector<Hypothesis_T> hypotheses_new;

	void reset_hypothesis(const Hypothesis_T & init);
	void update(void); ///<hypothesis becomes hypothesis_new, hypothesis_new becomes void
      };

#include "hypothesis_filter.tcc"
  }
}
#endif



// !!!put some default parametre and typedef a few hypothesis tracker 




    
