#ifndef IMAGE_GSTREAMERINPUT_H
#define IMAGE_GSTREAMERINPUT_H

#include "mimasconfig.h"

#if defined(HAVE_SYS_SEM_H) && defined(HAVE_SYS_SHM_H)

#include "image_input.h"

namespace mimas {

/** @addtogroup imageIO
    @{ */
/** @addtogroup imageInput
    @{ */
/** Class for reading image from gstreamer-input.
    <A HREF="http://gstreamer.freedesktop.org/">gstreamer</A> is an ongoing
    project, which will provide programmers with access to arbitrary video
    sources under GNU/Linux. <EM>mimas</EM> comes with a gstreamer-plugin
    (<EM>mimasimagesink</EM>) and this class, which provides shared memory
    access to video-frames obtained by <EM>gstreamer</EM>.

    Run <B>examples/gstreamer/runMe</B> to test gstreamer-input. */
template< typename T >
class image_gstreamerinput: public image_input< T >
{
 public:
  /** Constructor.
      The constructor requires a shared-memory-key and a semaphore-key to
      access the images provided by the plugin <EM>mimasimagesink</EM>.
      If the shared resources are not existing, an exception is issued.
      @param _shmkey Key to access shared memory.
      @param _semkey Key to access semaphores for synchronisation. */
  image_gstreamerinput( int _shmkey, int _semkey ) throw(mimasexception);
  /// Destructor.
  virtual ~image_gstreamerinput(void);
  /** Read image from shared memory.
      The image is read from shared memory. The semaphores are updated to
      allow the plugin to collect a new video-frame.
      @param image Object to store image in. */
  virtual void read( image< T > &image ) throw (mimasexception);  
 protected:
  /// Key to access semaphores.
  int semkey;
  /// Key to access shared memory.
  int shmkey;
  /// Id of semaphores.
  int semid;
  /// Id of shared memory.
  int shmid;
  /// Pointer to shared memory.
  void *shmem;
};

template< typename T, typename U >
inline image_gstreamerinput< T > &operator>>
   ( image_gstreamerinput< T > &input, image< U > &im )
{
  if ( typeid( T ) == typeid( U ) )
    input.read( im );
  else {
    image< T > temp;
    input.read( temp );
    im = image< U >( temp );
  };
  return input;
}

///@}

///@}

};

#include "image_gstreamerinput.tcc"

#endif

#endif

