#ifndef IMAGE_OUTPUT_H
#define IMAGE_OUTPUT_H

#include "image.h"

namespace mimas {

/** @addtogroup imageIO
    @{ */
/** @defgroup imageOutput Image Output
    This group contains classes for writing image(s) to different output sinks.
    @{ */
/** Abstract base-class for output sinks.
    This is a base-class for output sinks. It allows to implement a program,
    which can write image(s) to an arbitrary output-sink. F.e.:
    \code
    using namespace mimas;
    ...
    void test( image_output< rgba< unsigned char > > &sink ) {
       image< rgba< unsigned char > > image;
       ...
       if ( sink << image ) {
          ...
       } else {
          // Output-sink has closed.
       }
      
       ...
    };
    ...
    \endcode */
template< typename T >
class image_output: public object
{
 public:
  /// Constructor.
   image_output(void): state(true) {}
  /** Write image to arbitrary sink.
      This function is virtual and has to be overriden by the inheriting class.
      @param image The image to write to the sink. */
  virtual void write( const image< T > &image ) = 0;
  /// Indicator for output-sink being alive.
  operator bool(void) const { return state; }
  /// Indicator for output-sink having terminated.
  bool operator!(void) const { return !state; }
 protected:
  /// State variable.
  bool state;
};

/// Stream-operator for arbitrary output-sink.
template< typename T >
inline image_output< T > &operator<<( image_output< T > &output,
                                         const image< T > &image )
{ output.write( image ); return output; }

///
struct _wait { int milliSeconds; };
/** Stream-manipulator for performing a wait.
    @param _milliSeconds Amount of milliseconds to wait. */
inline _wait wait( int _milliSeconds )
{ _wait retVal; retVal.milliSeconds = _milliSeconds; return retVal; }

///
struct _pause {};
/** Stream-manipulator for pausing.
    Wait for user to close window. */
inline _pause pause(void)
{ _pause retVal; return retVal; }

///
struct _delay { int milliSeconds; };
/** Stream-manipulator for adjusting the frame-delay.
    This operator will adjust the framerate. */
inline _delay delay( int _milliSeconds )
{ _delay retVal; retVal.milliSeconds = _milliSeconds; return retVal; }

///@}

///@}

};

#endif
