#ifndef __LSYSTEM_H
#define __LSYSTEM_H


#include <boost/algorithm/string.hpp>
#include <string>
#include <list>
#include "image.h"
#include "image_draw.h"

namespace mimas {

/** @defgroup lsystem L-System
    Mimas-API for generating L-Systems
    @{ */
/** L-Systems.
    Also see
    <A HREF="http://algorithmicbotany.org/vmm-deluxe/Section-08.html">algorithmic botanic</A> by Przemyslaw Prusinkiewicz.
    @author Bala Amavasai (bala@amavasai.org)
    @date Thu Dec 15 17:05:19 GMT 2005 */
	class lsystem: public object
	{
		public:
			void setLstring(std::string str);
			std::string getLstring();
			void setAxiom(std::string str); ///< The axiom of the L-system. This is set to "F" as the default and seldom changes
			void setRule(std::string rule); ///< The rule for the L-system
			void setAngleStep(int val); ///< The step angles used when drawing the L-system
			void iterate(int numIterations=1); ///< The number of iterations
			void setLineLength(double val); ///< The line lengths of each component when drawing
			template<typename T>void drawLsystem(image<T> &image, int startx, int starty, T colour); ///< draw the L-system into an image
			lsystem();

		private:
			std::string lstring, axiom, rule;
			int anglestep;
			double d;
	};



	lsystem::lsystem()
	{
		// Example L-system
		// Adrian Mariono's Bush
		setAxiom("F");
		setRule("FF-[-F+F+F]+[+F-F-F]");
		setLstring("++++F");
		setAngleStep(22);
		setLineLength(5.0);
	}


	void lsystem::setLstring(std::string str)
	{
		lstring=str;
	}


	std::string lsystem::getLstring()
	{
		return lstring;
	}


	void lsystem::setAxiom(std::string str)
	{
		axiom=str;
	}


	void lsystem::setRule(std::string str)
	{
		rule=str;
	}



	void lsystem::setAngleStep(int val)
	{
		anglestep=val;
	}


	void lsystem::iterate(int numIterations)
	{
		if (numIterations<1) return;

		for (int i=0; i<numIterations; i++)
			boost::replace_all(lstring,axiom,rule);
	}


	void lsystem::setLineLength(double val)
	{
		// distance to move when drawing
		d=val;
	}

	template<typename T>
		void lsystem::drawLsystem(image<T> &image, int startx, int starty, T colour)
		{
			std::list<double> xlist;
			std::list<double> ylist;
			std::list<int> anglelist;

			double x=(double)startx,y=(double)starty;
			int angle=0; // initial angle
			double theta;

			for (int i=0; i<(int)lstring.length(); i++)
			{
				double newx, newy;
				theta=M_PI*(double)angle/180.0;
				switch(lstring[i])
				{
					case 'F' :
						newx=(x+d*cos(theta));
						newy=(y-d*sin(theta));
						drawLine(image,(int)x,(int)y,(int)newx,(int)newy,colour);
						x=newx;
						y=newy;
						break;
					case '+' :
						angle+=anglestep;
						break;
					case '-' : 
						angle-=anglestep;
						break;
					case '[':
						xlist.push_back(x);
						ylist.push_back(y);
						anglelist.push_back(angle);
						break;
					case ']':
						if (anglelist.size()>0)
						{
							x=xlist.back();
							y=ylist.back();
							angle=anglelist.back();
							xlist.pop_back();
							ylist.pop_back();
							anglelist.pop_back();
						}
						break;
					default:
						break;
				}

			}

		}

	///@}

};


#endif
