#ifndef NLINALG_H
#define NLINALG_H

#include <boost/numeric/ublas/vector.hpp>
#include "mimasconfig.h"

#ifdef HAVE_LIBMINPACK

namespace mimas {

/** @defgroup minpack Minpack functions
    Wrappers or bindings for Minpack functions are provided. The
    vector-classes of boost are used as datatypes.

    Read <A HREF="http://devernay.free.fr/hacks/netlibs.html">on how to install
    MINPACK</A>, if it is not included in your distribution.

    @author Julien Faucher (faucherj@gmail.com)
    @{ */

/** Function-object for lmdif.

    For more infomation about the algorithm read :
    <a href="http://www.tu-berlin.de/zrz/software/numerik/minpack/lmdif.txt">
    lmdif.txt</a>

    @see lmdif_ */
struct lmdif_t
{
  ///
  typedef boost::numeric::ublas::vector< double > Vector;

  /// Invoke minpack method.
  Vector operator()( const Vector &_A,
		     void (*fnc)(int *m, int *n, double *x, double *fvec,
			 int *iflag),
                     int nb, int maxfev, double tolerance, int mode, int factor, 
		     double epsfcn) const;
};

/** 
 * Levenberg-Marquardt algorithm. (Full version)
 *
 * @param A a vector containing a first approximation
 * @param fnc is the user-supplied subroutine which calculates the values of the 
 *        functions in a special point
 * @param nb the number of functions
 * @param maxfev the number of iterations to do
 * @param tolerance termination occurs when both the actual and predicted relative 
 *        reductions in the sum of squares are at most tol.
 * @param mode is an integer input variable. If mode = 1, the variables will be
 *        scaled internally. If mode = 2, the scaling is specified by the input diag.
 * @param factor is a positive input variable used in determining the initial
 *        step bound.
 * @param epsfcn is an input variable used in determining a suitable step length
 *        for the forward-difference approximation.
 *
 * @see lmdif_ for more information on the parameters (especially on fnc)
 */
inline 
lmdif_t::Vector lmdif ( const lmdif_t::Vector &A,
			void (*fnc)(int *m, int *n, double *x, 
				    double *fvec, int *iflag),
			int nb,
			int maxfev,
			double tolerance = 1e-7, 
			int mode = 1, 
			int factor = 1, 
			double epsfcn = 0)
{ 
  return lmdif_t()( A, fnc, nb, maxfev, tolerance, mode, factor, epsfcn ); 
}


/** Function-object for lmdif1.
    @see lmdif1_ */
struct lmdif1_t
{
  ///
  typedef boost::numeric::ublas::vector< double > Vector;

  /// Invoke minpack method.
  Vector operator()( const Vector &_A,
		     void (*fnc)(int *m, int *n, double *x, double *fvec,
			 int *iflag),
                     int nb, double tolerance) const;
};

/** 
 * Levenberg-Marquardt algorithm. (Simplified version)
 *
 * @param A a vector containing a first approximation
 * @param fnc is the user-supplied subroutine which calculates the values of the 
 *        functions in a special point
 * @param nb the number of functions
 * @param tolerance termination occurs when both the actual and predicted relative 
 *        reductions in the sum of squares are at most tol.
 *
 * @see lmdif_ for more information on the parameters (especially on fnc)
 */
inline 
lmdif1_t::Vector lmdif1 ( const lmdif1_t::Vector &A,
			  void (*fnc)(int *m, int *n, double *x, 
				      double *fvec, int *iflag),
			  int nb, 
			  double tolerance = 1e-7)
{ 
  return lmdif1_t()( A, fnc, nb, tolerance ); 
}

/** @} */

}

#endif

#endif

