#include "corner_detection.h"

#ifdef HAVE_LIBMINPACK
#ifdef CAMERA_CALIBRATION_ENABLE

namespace mimas {
namespace camera_calibration {

  /**
   * Initializes the template of the cross for the correlation algorithm
   */
  corner_detection::corner_detection(int size_x, int size_y, int size_window,
				     pixel_rgba white, pixel_rgba black) {

    // the values of the cross template for the 4 corners.
    // 1 : up left
    // 2 : up right
    // 3 : down right
    // 4 : down left
    const unsigned char values[][21*3] =
      { 
	{ 1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2, 3,
	  1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2, 3,
	  4, 5, 6, 10, 11, 12,
	  7, 8, 9,
	  10, 11, 12, 13, 14, 15,
	  1, 2, 3, 16, 17, 18, 1, 2, 3, 16, 17, 18,
	  1, 2, 3, 16, 17, 18, 1, 2, 3, 16, 17, 18 },

	{ 1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2, 3,
	  1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2, 3,
	  10, 11, 12, 4, 5, 6,
	  7, 8, 9,
	  13, 14 ,15, 10, 11, 12, 
	  16, 17, 18, 1, 2, 3, 16, 17, 18, 1, 2, 3,
	  16, 17, 18, 1, 2, 3, 16, 17, 18, 1, 2, 3 },

	{ 16, 17, 18, 1, 2, 3, 16, 17, 18, 1, 2, 3,
	  16, 17, 18, 1, 2, 3, 16, 17, 18, 1, 2, 3,
	  13, 14 ,15, 10, 11, 12, 
	  7, 8, 9,
	  10, 11, 12, 4, 5, 
	  1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2, 3,
	  1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2, 3 },

	{ 1, 2, 3, 16, 17, 18, 1, 2, 3, 16, 17, 18,
	  1, 2, 3, 16, 17, 18, 1, 2, 3, 16, 17, 18,
	  10, 11, 12, 13, 14, 15,
	  7, 8, 9,
	  4, 5, 6, 10, 11, 12,
	  1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2, 3,
	  1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2, 3 }
      };

    for(int i = 0; i < 4 ; ++i)
      for(int j = 0; j < 21*3 ; ++j) {
	switch(values[i][j]) {
	case 1: 
	  cross[i].push_back(white.getRed()); 
	  break;
	case 2: 
	  cross[i].push_back(white.getGreen()); 
	  break;
	case 3: 
	  cross[i].push_back(white.getBlue());
	  break;
        case 4: 
	  cross[i].push_back(white.getRed() - 10);
	  break;
        case 5:
	  cross[i].push_back(white.getGreen() - 10);
	  break;
        case 6:
	  cross[i].push_back(white.getBlue() - 10);
	  break;
        case 7:
	  cross[i].push_back((white.getRed() + black.getRed()) / 2);
	  break;
        case 8:
	  cross[i].push_back((white.getGreen() + black.getGreen()) / 2);
	  break;
        case 9:
	  cross[i].push_back((white.getBlue() + black.getBlue()) / 2);
	  break;
        case 10:
	  cross[i].push_back((white.getRed() + black.getRed()) / 2 + 10);
	  break;
        case 11:
	  cross[i].push_back((white.getGreen() + black.getGreen()) / 2 + 10);
	  break;
        case 12:
	  cross[i].push_back((white.getBlue() + black.getBlue()) / 2 + 10);
	  break;
        case 13:
	  cross[i].push_back(black.getRed() + 10);
	  break;
        case 14:
	  cross[i].push_back(black.getGreen() + 10);
	  break;
        case 15:
	  cross[i].push_back(black.getBlue() + 10);
	  break;
        case 16:
	  cross[i].push_back(black.getRed());
	  break;
        case 17:
	  cross[i].push_back(black.getGreen());
	  break;
        case 18:
	  cross[i].push_back(black.getBlue());
	  break;
	}
      }
    
    this->size_x = size_x;
    this->size_y = size_y;
    this->size_window = size_window;

  }

  /**
   * Destructor
   */ 
  corner_detection::~corner_detection(void) {

    for (int i = 0; i < 4 ; ++i)
      cross[i].clear();

  }

  /**
   * Subpixel detection using the Blais and Rioux detector of the
   * fourth order.
   */
  void
  corner_detection::g4(int x, int y, double &xx, double &yy, Matrix &coeffs) {

    double gx, gx1, gy, gy1;

    gx = coeffs(x-2,y) + coeffs(x-1,y) - coeffs(x+1,y) - coeffs(x+2,y);
    
    if(coeffs(x-1,y) < coeffs(x+1,y)) {

      gx1 = coeffs(x-1,y) + coeffs(x,y) - coeffs(x+2,y) - coeffs(x+3,y);
      xx = gx / (gx - gx1); 

    } else {

      gx1 = coeffs(x-3,y) + coeffs(x-2,y) - coeffs(x,y) - coeffs(x+1,y);
      xx = gx1 / (gx1 - gx) - 1; 

    }

    gy = coeffs(x,y-2) + coeffs(x,y-1) - coeffs(x,y+1) - coeffs(x,y+2);
    
    if(coeffs(x,y-1) < coeffs(x,y+1)) {

      gy1 = coeffs(x,y-1) + coeffs(x,y) - coeffs(x,y+2) - coeffs(x,y+3);
      yy = gy / (gy - gy1); 

    } else {

      gy1 = coeffs(x,y-3) + coeffs(x,y-2) - coeffs(x,y) - coeffs(x,y+1);
      yy = gy1 / (gy1 - gy) - 1; 

    }

  }


  /**
   * Implementation of the template algorithm to find the corner.
   */
  pixel_f
  corner_detection::findLocalCorner(image< pixel_rgba > img, int x, int y, int precision, int type_corner, 
			   double &corr, double min_corr) throw(mimasexception) {

    std::vector< unsigned char > image;
    int im = 0, jm = 0, X, Y;

    // init the correlation coefficient to the minimum
    corr = -1;

    X = img.getWidth();
    Y = img.getHeight();

    MMERROR(x > 5 && y > 5 && x < X-5 && y < Y-5, mimasexception, ,
	    "Too close to the edge." );

    Matrix coeffs(1 - std::max(5,x-precision) + std::min(x+precision, X-6),
		  1 - std::max(5,y-precision) + std::min(y+precision, Y-6));

    for(int i = std::max(5,x-precision); i <= std::min(x+precision, X-6); ++i)
      for(int j = std::max(5,y-precision); j <= std::min(y+precision, Y-6); ++j) {

	// save the pixels corresponding with the template i.e. a cross
	// around the pixel to correlate
	image.clear();
	for(int k = -5; k <= 5; ++k) {

	  image.push_back(img.getPixel(i-std::abs(k),j+k).getRed());
	  image.push_back(img.getPixel(i-std::abs(k),j+k).getGreen());
	  image.push_back(img.getPixel(i-std::abs(k),j+k).getBlue());

	  if(k != 0) {
	    image.push_back(img.getPixel(i+std::abs(k),j+k).getRed());
	    image.push_back(img.getPixel(i+std::abs(k),j+k).getGreen());
	    image.push_back(img.getPixel(i+std::abs(k),j+k).getBlue());

	  }

	}
	
	// correlate with the template
	corrcoef< std::vector< unsigned char > > c;

	// save the coefficient
	coeffs(i-std::max(5,x-precision), j-std::max(5,y-precision)) = c(image, cross[type_corner]);

	// save the best correlation
	if(coeffs(i-std::max(5,x-precision), j-std::max(5,y-precision)) > corr) {
	  corr = coeffs(i-std::max(5,x-precision), j-std::max(5,y-precision));
	  im = i;
	  jm = j;
	}      

      }

    // test that the correlation is greater than the minimum required
    MMERROR(corr >= min_corr , mimasexception, , "Correlation too small." );

    bool test = true;

    if(im < x) {
      test &= (x - im) < (x - std::max(5,x-precision) - 3);
    }
    else{
      test &= (im - x) < (std::min(x+precision, X-6) - x - 3);
    }

    if(jm < y) {
      test &= (y - jm) < (y - std::max(5,y-precision) - 3);
    }
    else{
      test &= (jm - y) < (std::min(y+precision, Y-6) - y - 3);
    }
    MMERROR(test, mimasexception, , "Bad approximation of the corner." );

    double xx, yy;

    g4(im-std::max(5,x-precision),jm-std::max(5,y-precision),xx,yy,coeffs);

    return pixel_f(im + xx, jm + yy );

  }


  /**
   * Finds the corners of an chessboard, given the 4 extreme corners.
   *
   * This function computes with the extrem corners an homography in order to
   * get a first approximation of the corner locations, and uses the detection
   * algorithm in a window around the approximed pixel, to get the accurate one.
   */
  void corner_detection::findCorners(image< pixel_rgba > img, std::vector< pixel_f > pix, 
			    double min_corr) throw(mimasexception) {

    MMERROR(pix.size() == 4, mimasexception, ,
	    "Bad size." );

    Chessboard c;

    c.image = pix;

    c.object.push_back(point_f(0, 0));
    c.object.push_back(point_f(2*size_x-1, 0));
    c.object.push_back(point_f(2*size_x-1, 2*size_y-1));
    c.object.push_back(point_f(0, 2*size_y-1));

    // find the homography to get a first approximation
    Camera::Matrix hg(3,3);
    Camera::findHomography(c, hg);

    MatrixPixel corners(2*size_y, 2*size_x);
    Camera::Vector v(3), r(3);

    v(2) = 1; // to be done only once

    double corr;

    // find all the corners
    for (unsigned i = 0; i < corners.size1 (); ++ i){
      for (unsigned j = 0; j < corners.size2 (); ++ j) {

	// choose the type of the corner depending on the parity of i and j
	int type = (i%2) ? ( (j%2) ? 2 : 3 ) : ( (j%2) ? 1 : 0 ); 

	v(0) = j;
	v(1) = i;

	// find a first approximation of the corner location, using the homography
	r = boost::numeric::ublas::prod(hg, v);

	// find the accurate location of the corner, using the dectection algorithm
	// and the first approximation
	corners(i, j) = findLocalCorner(img, (int)(r(0)/r(2)), (int)(r(1)/r(2)),
					size_window, type, corr, min_corr);

      }
    }

    // save the corners
    mCorners = corners;

  }

}
}

#endif
#endif
