#include <errno.h>
#include <Magick++.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include "gnuplot.h"

using namespace std;

namespace mimas {

pid_t gnuplot::pid = 0;

void gnuplot::alarmHandler( int signal )
{
  kill( gnuplot::pid, SIGQUIT );
}

gnuplot::gnuplot( image< rgba< unsigned char > > &_img, int width, int height,
                    int _timeout )
  throw(mimasexception):
  img(&_img), timeout(_timeout), finished(false)
{

  MMERROR( pipe( pipes ) == 0, mimasexception, ,
           "Failed to create pipes for connecting to 'gnuplot': "
           << strerror( errno ) );

  assert( pid == 0 );
  pid = fork();

  if ( pid < 0 ) {

    close( pipes[0] );
    close( pipes[1] );
    MMERROR( false, mimasexception, ,
             "Failed to create process for 'gnuplot': "
             << strerror( errno ) );

  } else if ( pid == 0 ) {

    // This is the new process for 'gnuplot'.
    dup2( pipes[0], STDIN_FILENO );
    dup2( pipes[1], STDOUT_FILENO );
    execlp( "/usr/bin/gnuplot", "gnuplot", NULL );
    
    // 'gnuplot' was not found.
    exit( EXIT_FAILURE );
    
  } else {
    
    // Main process continues here.
    program << "set terminal png medium size " << width << "," << height
            << endl;
    
  };

};

gnuplot::~gnuplot(void)
{
  if ( !finished ) {
    try {
      finish();
    } catch ( mimasexception & ) {
    };
  };
  close( pipes[0] );
  close( pipes[1] );
}

void gnuplot::finish(void) throw (mimasexception)
{
  if ( !finished ) {

    // Do not try this twice!
    finished = true;

    program << endl << "quit" << endl;
    int woffset = 0;
    while ( woffset < (signed)program.str().length() ) {
      ssize_t written =
        write( pipes[1], program.str().c_str() + woffset,
               program.str().length() - woffset );
      MMERROR( written >= 0, mimasexception, ,
               "Error writing to input-pipe of 'gnuplot': "
               << strerror( errno ) );
      woffset += written;
    };

    sighandler_t oldHandler = signal( SIGALRM, alarmHandler );
    alarm( timeout );
    int status;
    bool result = ( waitpid( pid, &status, 0 ) == pid );
    alarm( 0 );
    signal( SIGALRM, oldHandler );
    pid = 0;

    MMERROR( result, mimasexception, ,
             "Error indicated by waitpid: " << strerror( errno ) );
    MMERROR( WIFEXITED(status), mimasexception, ,
             "'gnuplot' did not exit normally." );
    MMERROR( WEXITSTATUS(status) == 0, mimasexception, ,
             "The return-value of 'gnuplot' is not zero (it is "
             << WEXITSTATUS(status) << ")" );
    
    int count;
    ioctl( pipes[0], FIONREAD, &count );
    
    MMERROR( count > 0, mimasexception, ,
             "'gnuplot' did not render a plot so far." );
      
    // 'gnuplot' returned some data.
    char *data = new char[ count ];
    Magick::Blob blob;
    blob.updateNoCopy( data, count, Magick::Blob::NewAllocator );

    int roffset = 0;
    while ( roffset < count ) {
      ssize_t readsize = read( pipes[0], data + roffset, count - roffset );
      MMERROR( readsize >= 0, mimasexception, ,
               "Error reading from output-pipe of 'gnuplot': "
               << strerror( errno ) );
      roffset += readsize;
    };
    
    std::list< Magick::Image > images;
    readImages( &images, blob );
    
    MMERROR( !images.empty(), mimasexception, ,
             "Did not recognise image-data returned by 'gnuplot'." );
    
    std::list< Magick::Image >::iterator i = images.begin();
    img->init( i->columns(), i->rows() );
    i->write( 0, 0, i->columns(), i->rows(), "BGRA", Magick::CharPixel,
              img->rawData().data() );
  };
}

void gnuplot::unsetLineTitle(void)
{
  lineTitles.clear();
}

void gnuplot::setLineTitle(const char *title, ...)
{
  unsetLineTitle();
	
  va_list args;
	
  va_start(args,title);

  // precount parameters
  int paramcount=-1;
  const char *val="dummy";
  while(val!=NULL) {
    paramcount++;
    val=va_arg(args,const char *);
  }
  paramcount++;

  lineTitles.push_back( title );

  va_end(args);

  va_start(args,title);

  // start from 1 since first parameter already loaded.
  for (int i=1; i<paramcount; i++) {
    val=va_arg(args,const char *);
    lineTitles.push_back( val );
  } 

  va_end(args);
}

void gnuplot::setMultiPlot(bool multiplot)
{
  program << "set " << ( multiplot ? "multiplot" : "nomultiplot" ) << endl;
}

}
