#include <boost/shared_array.hpp>
#include <errno.h>
#include "image_v4linput.h"

#ifdef HAVE_LINUX_VIDEODEV_H

#include "colourspace.h"

using namespace boost;
using namespace std;

namespace mimas {


template<>
void image_v4linput< rgba< unsigned char > >::read
  (image< rgba< unsigned char > > &img ) throw (mimasexception)
{
  // State-variable never becomes false. Input from video4linux only comes
  // to an end by an error.
  assert( state == true );

  shared_array< unsigned char > buffer;
  void *src = NULL;
  
  if ( map != MAP_FAILED ) {
    video_mmap vmap;
    vmap.frame = 0;
    vmap.width = win.width;
    vmap.height = win.height;
    vmap.format = pic.palette;

    MMERROR( xioctl( VIDIOCMCAPTURE, &vmap ) == 0, mimasexception, ,
             "Error initiating capture of image: " << strerror( errno ) );

    MMERROR( xioctl( VIDIOCSYNC, &vmap ) == 0, mimasexception, ,
             "Error on synchronising with capture of image: "
	     << strerror( errno ) );

    src = map;
  } else {
    int size;
    switch ( pic.palette ) {
    case VIDEO_PALETTE_RGB24:
      size = 3 * win.width * win.height;
      break;
    case VIDEO_PALETTE_YUV420P:
      size = win.width * win.height * 3 / 2;
      break;
    default:
      MMERROR( false, mimasexception, ,
               "Colour-grabbing with video-palette number " << pic.palette
               << " not supported by Mimas. Please contact the developers." );
    };
    buffer = shared_array< unsigned char >( new unsigned char[ size ] );
    src = (void *)buffer.get();
    MMERROR( ::read( fd, src, size ) != -1, mimasexception, ,
             "Error reading from device: " << strerror( errno ) );
  };

  img.init( win.width, win.height );

#ifndef NDEBUG
  std::cerr << img.rawData().data() << ' '
            << img.rawData().num_elements() << ' '
            << img.getWidth() << ' ' << img.getHeight() << std::endl;

#endif
  switch ( pic.palette ) {
  case VIDEO_PALETTE_RGB24:
    rgb_to_rgba( (const char *)src, win.width, win.height,
                 (char *)img.rawData().data() );
    break;
  case VIDEO_PALETTE_YUV420P:
    yuv420p_to_rgba( (const char *)src, win.width, win.height,
                     (char *)img.rawData().data() );
    break;
  default:
    MMERROR( false, mimasexception, ,
             "Colour-grabbing with video-palette number " << pic.palette
             << " not supported by Mimas. Please contact the developers." );
  };
}

template<>
void image_v4linput< unsigned char >::read
  (image< unsigned char > &img ) throw (mimasexception)
{
  // State-variable never becomes false. Input from video4linux only comes
  // to an end by an error.
  assert( state == true );

  assert( pic.palette = VIDEO_PALETTE_GREY );
  
  shared_array< unsigned char > buffer;
  void *src = NULL;
  int size = -1;
  img.init( win.width, win.height );

  if ( map != MAP_FAILED ) {
    video_mmap vmap;
    vmap.frame = 0;
    vmap.width = win.width;
    vmap.height = win.height;
    vmap.format = pic.palette;

    MMERROR( xioctl( VIDIOCMCAPTURE, &vmap ) == 0, mimasexception, ,
             "Error initiating capture of image: " << strerror( errno ) );

    MMERROR( xioctl( VIDIOCSYNC, &vmap ) == 0, mimasexception, ,
             "Error on synchronising with capture of image: "
	     << strerror( errno ) );

    src = map;
  } else {
    switch ( pic.palette ) {
    case VIDEO_PALETTE_GREY:
      size = win.width * win.height;
      src = img.rawData().data();
      break;
    case VIDEO_PALETTE_YUV420P:
      size = win.width * win.height * 3 / 2;
      buffer = shared_array< unsigned char >( new unsigned char[ size ] );
      src = (void *)buffer.get();
      break;
    default:
      MMERROR( false, mimasexception, ,
               "Greyscale-grabbing with video-palette number " << pic.palette
               << " not supported by Mimas. Please contact the developers." );
    };
    MMERROR( ::read( fd, src, size ) != -1, mimasexception, ,
             "Error reading from device: " << strerror( errno ) );
  };

#ifndef NDEBUG
  std::cerr << img.rawData().data() << ' '
            << img.rawData().num_elements() << ' '
            << img.getWidth() << ' ' << img.getHeight() << std::endl;

#endif
  switch ( pic.palette ) {
  case VIDEO_PALETTE_GREY:
    if ( map != MAP_FAILED ) {
      assert( size != -1 );
      memcpy( img.rawData().data(), map, size );
    };
    break;
  case VIDEO_PALETTE_YUV420P:
    yuv420p_to_grey( (const char *)src, win.width, win.height,
                     (char *)img.rawData().data() );
    break;
  default:
    MMERROR( false, mimasexception, ,
             "Greyscale-grabbing with video-palette number " << pic.palette
             << " not supported by Mimas. Please contact the developers." );
  };
}

template<>
void image_v4linput< rgba< unsigned char > >::selectPalette(void)
  throw(mimasexception)
{
  boost::array< __u16, 14 > palette;
  palette[  0 ] = VIDEO_PALETTE_RGB24;// ->RGBA supported
  palette[  1 ] = VIDEO_PALETTE_YUV420P;// ->RGBA supported
  palette[  2 ] = VIDEO_PALETTE_HI240;
  palette[  3 ] = VIDEO_PALETTE_RGB565;
  palette[  4 ] = VIDEO_PALETTE_RGB555;
  palette[  5 ] = VIDEO_PALETTE_RGB32;
  palette[  6 ] = VIDEO_PALETTE_YUV422;
  palette[  7 ] = VIDEO_PALETTE_YUYV;
  palette[  8 ] = VIDEO_PALETTE_UYVY;
  palette[  9 ] = VIDEO_PALETTE_YUV420;
  palette[ 10 ] = VIDEO_PALETTE_YUV411;
  palette[ 11 ] = VIDEO_PALETTE_RAW;
  palette[ 12 ] = VIDEO_PALETTE_YUV422P;
  palette[ 13 ] = VIDEO_PALETTE_YUV411P;
  
  pic.brightness = pic.hue = pic.colour = pic.contrast = pic.whiteness  = 32767;
  pic.depth = 24;
  
  int selected = 0;
    
  while ( true ) {
	  
    pic.palette = palette[ selected ];
    int r = xioctl( VIDIOCSPICT, &pic );
    if ( r == 0 ) {
#ifndef NDEBUG
      std::cerr << "Using " << ( selected + 1 ) << "th supported option."
                << std::endl;
#endif
      break;
    };
    selected++;
    MMERROR( selected < (signed)palette.size(),
             mimasexception, ,
             "Camera-driver doesn't offer a colour-video-palette known to "
             "mimas." );
    
  };
}

template<>
void image_v4linput< unsigned char >::selectPalette(void)
  throw(mimasexception)
{
  boost::array<  pair< __u16, __u16 >, 14 > palette;
  palette[  0 ] = pair< __u16, __u16 >( VIDEO_PALETTE_GREY   ,  8 );//support
  palette[  1 ] = pair< __u16, __u16 >( VIDEO_PALETTE_YUV420P, 24 );//supported
  palette[  2 ] = pair< __u16, __u16 >( VIDEO_PALETTE_RGB24  , 24 );
  palette[  3 ] = pair< __u16, __u16 >( VIDEO_PALETTE_HI240  , 24 );
  palette[  4 ] = pair< __u16, __u16 >( VIDEO_PALETTE_RGB565 , 24 );
  palette[  5 ] = pair< __u16, __u16 >( VIDEO_PALETTE_RGB555 , 24 );
  palette[  6 ] = pair< __u16, __u16 >( VIDEO_PALETTE_RGB32  , 24 );
  palette[  7 ] = pair< __u16, __u16 >( VIDEO_PALETTE_YUV422 , 24 );
  palette[  8 ] = pair< __u16, __u16 >( VIDEO_PALETTE_YUYV   , 24 );
  palette[  9 ] = pair< __u16, __u16 >( VIDEO_PALETTE_UYVY   , 24 );
  palette[ 10 ] = pair< __u16, __u16 >( VIDEO_PALETTE_YUV420 , 24 );
  palette[ 11 ] = pair< __u16, __u16 >( VIDEO_PALETTE_YUV411 , 24 );
  palette[ 12 ] = pair< __u16, __u16 >( VIDEO_PALETTE_RAW    , 24 );
  palette[ 13 ] = pair< __u16, __u16 >( VIDEO_PALETTE_YUV422P, 24 );
  palette[ 14 ] = pair< __u16, __u16 >( VIDEO_PALETTE_YUV411P, 24 );
  
  pic.brightness = pic.hue = pic.colour = pic.contrast = pic.whiteness  = 32767;
  
  int selected = 0;
    
  while ( true ) {
	  
    pic.palette = palette[ selected ].first;
    pic.depth   = palette[ selected ].second;
    int r = xioctl( VIDIOCSPICT, &pic );
    if ( r == 0 ) {
#ifndef NDEBUG
      std::cerr << "Using " << ( selected + 1 ) << "th supported option."
                << std::endl;
#endif
      break;
    };
    selected++;
    MMERROR( selected < (signed)palette.size(),
             mimasexception, ,
             "Camera-driver doesn't offer a video-palette known to "
             "mimas." );
  };
}

};
#endif

