//
// Multi-layered perceptron tool class
//
// Bala Amavasai (bala@amavasai.org)
// $Header: /cvs/mimas2/lib/nn_mlp_tool.cc,v 1.1.1.1 2005/08/09 15:37:29 engmb Exp $
//

#include <boost/random/linear_congruential.hpp>
#include <boost/random/normal_distribution.hpp>
#include <boost/random/variate_generator.hpp>
#include <sys/types.h>
#include <fstream>
#include <sys/time.h>
#include "nn_mlp_tool.h"

using namespace boost;
using namespace boost::random;
using namespace std;

namespace mimas {

nn_mlp_tool::nn_mlp_tool(void)
{
	network_initialised=false;
	debugtrain=false;
	//learningRate=0.3;
	//momentum=0.05;
	iterations=1000;


	// weightDecay
	// weightDecay=0.0005; // Online
	// weightDecay=0.0005; // Batch
	// weightDecay=; // QProp
	// weightDecay=0.001; // RProp
}
	
nn_mlp_tool::~nn_mlp_tool(void)
{
	clear();
}


// clear the dang network
void nn_mlp_tool::clear(void)
{
	if (network_initialised)
	{
     weight.reset();
     bias.reset();
     y.reset();
     net.reset();
     delta.reset();
     prevWeightChange.reset();
     prevBiasChange.reset();
     numnodes.reset();

     ptrToTransferFunc.reset();
     ptrToTransferFuncDeriv.reset();

	}
	network_initialised=false;
}


void nn_mlp_tool::showConfig(void)
{
	cerr << endl;
	cerr << "Results of nn_mlp_tool::showConfig" << endl;
	cerr << "====================================" << endl;
	
	if (!network_initialised)
	{
		cerr << "Network not initialised!" << endl;
		return;
	}

	cerr << "Network configuration is :  " ;
	for (int l=0; l<numlayers-1; l++)
		cerr << numnodes[l] << "-";

	cerr << numnodes[numlayers-1] << endl;

	cerr << endl;
}


// private function which initialises network based on numnodes[]
void nn_mlp_tool::init(void)
{
	// initialise matrices and arrays
  weight=shared_array< Matrix >( new Matrix[numlayers] );
  bias=shared_array< Vector >( new Vector[numlayers] );
  y=shared_array< Vector >( new Vector[numlayers] );
  net=shared_array< Vector >( new Vector[numlayers] );
  delta=shared_array< Vector >( new Vector[numlayers] );
  prevWeightChange=shared_array< Matrix >( new Matrix[numlayers] );
  prevBiasChange=shared_array< Vector >( new Vector[numlayers] );

	// initialise pointers to transfer functions and derivatives
	ptrToTransferFunc=shared_array< ptrToFuncType >( new ptrToFuncType[numlayers] );
	ptrToTransferFuncDeriv=shared_array< ptrToFuncType >( new ptrToFuncType[numlayers] );

   // put some random values in the weight and bias matricesa

   // Create random-generator with normal distribution (with zero mean and
   // 1.0 standard deviation).
   rand48 gen( (uint64_t)time( NULL ) );
    variate_generator< rand48, normal_distribution< double > >
       n( gen,
        normal_distribution< double >( 0.0, 1.0 ) );

	// put some random values in the weight and bias matrices
	for (int i=1; i<numlayers; i++)
	{
     weight[ i ].resize( numnodes[i], numnodes[i-1] );
     generate( weight[ i ].data().begin(), weight[ i ].data().end(), n );

     prevWeightChange[ i ].resize( numnodes[i], numnodes[i-1] );

     bias[ i ].resize( numnodes[i] );
     generate( bias[ i ].data().begin(), bias[ i ].data().end(), n );

     prevBiasChange[ i ].resize( numnodes[i] );

     ptrToTransferFunc[ i ] = &nn_mlp_tool::transferFuncTanh;
     ptrToTransferFuncDeriv[ i ] = &nn_mlp_tool::transferFuncDerivTanh;
	}

	for (int i=0; i<numlayers; i++)
	{
		y[i].resize(numnodes[i]);
		net[i].resize(numnodes[i]);
		delta[i].resize(numnodes[i]);
	}

	network_initialised=true;
}


void nn_mlp_tool::init(int nodes, ...) throw (mimasexception)
{
	int nlayers=0;

	va_list args;
	va_start(args,nodes);

	//precount parameters until first NULL
	int val=-1;
	while(val!=0)
	{
		nlayers++;
		val=va_arg(args,int);
	}
	va_end(args);
	
   MMERROR( nlayers>=3, mimasexception, , "nn_mlp_tool::init - requires at "
            "least 3 layers: 1 input, 1 hidden and 1 output" );

	// initialise some vals
	numlayers=nlayers;
	numnodes=shared_array< int >( new int[numlayers] );

	// get number of nodes in each layer
	va_start(args, nodes);

	numnodes[0]=nodes;
	for (int i=1; i<numlayers; i++)
		numnodes[i]=va_arg(args, int);
	
	va_end(args);
	init();

}

// load neural network from file
void nn_mlp_tool::load (const char *fn) throw (mimasexception)
{
	ifstream infile(fn);

	if (network_initialised) clear();
		
   MMERROR( infile.is_open(), mimasexception, ,
            "nn_mlp_tool::load - file cannot be opened" );

	MMERROR( infile >> numlayers, mimasexception, ,
            "nn_mlp_tool::load - unable to load number of layers" );

	numnodes=shared_array< int >( new int[numlayers] );
	for (int i=0; i<numlayers; i++)
		infile >> numnodes[i];

	init(); // initialise network based on numnodes[];

	// save the weights and biases
	for (int i=1; i<numlayers; i++)
	{
		infile >> weight[i];
		infile >> bias[i];
	}

	infile.close();
}



// save neural network to a file
void nn_mlp_tool::save (const char *fn) throw (mimasexception)
{
	ofstream outfile(fn);
	
   MMERROR( network_initialised, mimasexception, ,
            "nn_mlp_tool::save - cannot save a network that are not "
            "initialised" );
	
	MMERROR( outfile.is_open(), mimasexception, ,
            "nn_mlp_tool::save - file cannot be opened" );

	// store number of layers followed by number of nodes in each layer
	outfile << numlayers << endl;
	for (int i=0; i<numlayers; i++)
		outfile << numnodes[i] << " ";
	outfile << endl;
	
	// save the weights and biases
	for (int i=1; i<numlayers; i++)
	{
		outfile << weight[i];
		outfile << bias[i];
	}

	outfile.close();

	
   }


// set number of iterations to train for
void nn_mlp_tool::setIterations(int i) throw (mimasexception)
{
  MMERROR( i>0, mimasexception, ,
           "nn_mlp_tool::setIterations - parameter cannot be <=0" );
	iterations=i;

}

// set number of iterations to train for
int nn_mlp_tool::getIterations(void)
{
	return iterations;
}


void nn_mlp_tool::setTransferFuncLayer(ptrToFuncType newfunc, int layer)
  throw (mimasexception)
{
  MMERROR( network_initialised && layer>=2 && layer<=numlayers, mimasexception, ,
           "nn_mlp_tool::setTransferFunc - requires at least 3 layers: "
           "1 input, 1 hidden and 1 output" );

	// layer-1 below since user is indexing from 1. The first hidden layer is 2
	ptrToTransferFunc[layer-1]=newfunc;
}

void nn_mlp_tool::setTransferFuncDerivLayer(ptrToFuncType newfunc, int layer)
  throw (mimasexception)
{
  MMERROR( network_initialised && layer>=2 && layer<=numlayers, mimasexception, ,
           "nn_mlp_tool::setTransferFunc - requires at least 3 layers: "
           "1 input, 1 hidden and 1 output" );

	// layer-1 below since user is indexing from 1. The first hidden layer is 2
	ptrToTransferFuncDeriv[layer-1]=newfunc;
}

double nn_mlp_tool::transferFunc(double val)
{
	return tanh(val);
}
	

double nn_mlp_tool::transferFuncDeriv(double val)
{
	double temp=tanh(val);
	return 1.0-(temp*temp);
}


double nn_mlp_tool::transferFuncTanh(double val)
{
	return tanh(val);
}
	

double nn_mlp_tool::transferFuncDerivTanh(double val)
{
	double temp=tanh(val);
	return 1.0-(temp*temp);
}

double nn_mlp_tool::transferFuncLinear(double val)
{
	return val;
}
	

double nn_mlp_tool::transferFuncDerivLinear(double val)
{
	return 1.0;
}



nn_mlp_tool::Matrix nn_mlp_tool::feedForward(Matrix& pat) throw (mimasexception)
{
   MMERROR( network_initialised && ( numnodes[0]==(signed)pat.size2()),
            mimasexception, ,
            "nn_mlp_tool::train - input/output pattern dimensions not "
            "consistent" );

   Matrix ffvals(pat.size1(),numnodes[numlayers-1]);
	
	// loop through all patterns
	for (int p=0; p<(signed)pat.size1(); p++)
	{

		// forward propagation on first layer
		for (int j=0; j<numnodes[0]; j++)
			y[0](j)=pat(p,j);
		// forward propagation on other layers
		for (int l=1; l<numlayers; l++)
		{
			//cerr << weight[l] << endl;
			//cerr << y[l-1] << endl;
			//cerr << bias[l] << endl;
        net[l] = prod( weight[l], y[l-1] ) + bias[l];

			// transfer function
			// for (int j=0; j<numnodes[l]; j++)
			//	y[l](j,0)=transferFunc(net[l](j,0));

			// transfer function
			for (int j=0; j<numnodes[l]; j++)
           y[l](j)= ( this->*ptrToTransferFunc[l] )( net[l](j) );

		}
		
		for (int j=0; j<numnodes[numlayers-1]; j++)
			ffvals(p,j)=y[numlayers-1](j);

	}

	return ffvals; // return feedforward values


}


// perform deltaRule on pattern on row p
double nn_mlp_tool::deltaRule(Matrix& pat, int p)
  throw (mimasexception)
{
	double sseval=0.0;
	
	// check that pat is initialised and the number of columns is equivalent
	// to the number of input nodes + output nodes
   MMERROR( network_initialised && p>=0 && p<(signed)pat.size1(),
            mimasexception, ,
            "nn_mlp_tool::deltaRule - input/output pattern dimensions not "
            "consistent or pattern number not in range" );

	// forward propagation on first layer
	for (int j=0; j<numnodes[0]; j++)
		y[0](j)=pat(p,j);

	// forward propagation on other layers
	for (int l=1; l<numlayers; l++)
	{
		//cerr << weight[l] << endl;
		//cerr << y[l-1] << endl;
		//cerr << bias[l] << endl;
     net[l] = prod( weight[l], y[l-1] ) + bias[l];

		// transfer function
		// for (int j=0; j<numnodes[l]; j++)
		//	y[l](j,0)=transferFunc(net[l](j,0));

		// transfer function
		for (int j=0; j<numnodes[l]; j++)
			y[l](j)=(this->*ptrToTransferFunc[l])(net[l](j));
	}

	// calculate error in output layer
	for (int j=0; j<numnodes[numlayers-1]; j++)
	{
		double err=(pat(p,numnodes[0]+j)-y[numlayers-1](j));

		//delta[numlayers-1](j,0)=transferFuncDeriv(net[numlayers-1](j,0))*err;
		delta[numlayers-1](j)=(this->*ptrToTransferFuncDeriv[numlayers-1])(net[numlayers-1](j))*err;
		sseval+=(err*err);
	}

	// backpropagate the errors for other layers
	for (int l=numlayers-2; l>=1; l--)
	{
		// delta[l]= ( weight[l+1]^T * delta[l+1] ) * f'[l](net[l]);


		// note: use matrix form in the future! ;)
		for (int j=0; j<numnodes[l]; j++)
		{
			delta[l](j)=0.0;

			for (int k=0; k<numnodes[l+1]; k++)
				delta[l](j)=delta[l](j)+delta[l+1](k)*weight[l+1](k,j);

			//delta[l](j,0)=delta[l](j,0)*transferFuncDeriv(net[l](j,0));
			delta[l](j)=delta[l](j)*(this->*ptrToTransferFuncDeriv[l])(net[l](j));
		}
	}


	return sseval;
}


nn_mlp_tool::Matrix nn_mlp_tool::trainOnline(Matrix& pat, double learningRate, double momentum, double weightDecay) throw (mimasexception)
{
	double sse;
	
	// check that pat is initialised and the number of columns is equivalent
	// to the number of input nodes + output nodes
   MMERROR( network_initialised &&
            (numnodes[0]+numnodes[numlayers-1]==(signed)pat.size2()),
            mimasexception, ,
            "nn_mlp_tool::train - input/output pattern dimensions not consistent" );

	Matrix ssevals(iterations,2);

	for (int iter=0; iter<iterations; iter++)
	{

		sse=0.0;

		// loop through all patterns
		for (int p=0; p<(signed)pat.size1(); p++)
		{
			sse+=deltaRule(pat,p);

			double weightupdate; // used to speed-up computation

			// update weights in all layers
			for (int l=1; l<numlayers; l++)
			{
				for (int k=0; k<numnodes[l]; k++)
				{
					for (int j=0; j<numnodes[l-1]; j++)
					{
						weightupdate=(learningRate*delta[l](k)*y[l-1](j))+
							(momentum*prevWeightChange[l](k,j))
							-(weightDecay*weight[l](k,j));

						// update weight
						weight[l](k,j)=weight[l](k,j)+weightupdate;

						// store weight updates
						prevWeightChange[l](k,j)=weightupdate;
					}

					weightupdate=(learningRate*delta[l](k)*1.0) + ( momentum * prevBiasChange[l](k)) - (weightDecay*1.0);
					bias[l](k)=bias[l](k) + weightupdate;
					prevBiasChange[l](k)=weightupdate;

				}
			}

		} // pat

		ssevals(iter,0)=(double)iter;
		ssevals(iter,1)=sse;

		if (debugtrain) cout << iter << " \t" << sse << endl;
	} // iter



	return ssevals;
}

nn_mlp_tool::Matrix nn_mlp_tool::trainBatch(Matrix& pat, double learningRate, double momentum, double weightDecay) throw (mimasexception)
{
	double sse;
	
	// check that pat is initialised and the number of columns is equivalent
	// to the number of input nodes + output nodes
	MMERROR( network_initialised &&
            (numnodes[0]+numnodes[numlayers-1]!=(signed)pat.size2()),
            mimasexception, ,
            "nn_mlp_tool::train - input/output pattern dimensions not "
            "consistent" );


	// initialise accumulators for batch training.
	shared_array< Matrix > accumWeightUpdate( new Matrix[numlayers] );
	shared_array< Vector > accumBiasUpdate( new Vector[numlayers] );

   Matrix ssevals(iterations,2);

	for (int iter=0; iter<iterations; iter++)
	{

		sse=0.0;
		
		// Zero accumulators
		for (int l=1; l<numlayers; l++)
		{
			accumWeightUpdate[l].resize(numnodes[l],numnodes[l-1]);
			accumBiasUpdate[l].resize(numnodes[l]);
		}


		// loop through all patterns
		for (int p=0; p<(signed)pat.size1(); p++)
		{
			sse+=deltaRule(pat,p);

			// accummulate deltas for batch backpropagation
			for (int l=1; l<numlayers; l++)
			{
				for (int k=0; k<numnodes[l]; k++)
				{
					for (int j=0; j<numnodes[l-1]; j++)
						accumWeightUpdate[l](k,j)=accumWeightUpdate[l](k,j)+(delta[l](k)*y[l-1](j));

					accumBiasUpdate[l](k)=accumBiasUpdate[l](k)+(delta[l](k)*1.0);
				}
			}


		} // pat


		// Perform weight update
		double weightupdate; // used to speed-up computation
		// update weights in all layers
		for (int l=1; l<numlayers; l++)
		{
			for (int k=0; k<numnodes[l]; k++)
			{
				for (int j=0; j<numnodes[l-1]; j++)
				{
					weightupdate=(learningRate*accumWeightUpdate[l](k,j)/pat.size1())+
						(momentum*prevWeightChange[l](k,j))
						-(weightDecay*weight[l](k,j));

					// update weight
					weight[l](k,j)=weight[l](k,j)+weightupdate;

					// store weight updates
					prevWeightChange[l](k,j)=weightupdate;
				}

				weightupdate=(learningRate*accumBiasUpdate[l](k)/pat.size1()) + ( momentum * prevBiasChange[l](k))-(weightDecay*1.0);
				bias[l](k)=bias[l](k) + weightupdate;
				prevBiasChange[l](k)=weightupdate;

			}
		}

		ssevals(iter,0)=(double)iter;
		ssevals(iter,1)=sse;

		if (debugtrain) cout << iter << " \t" << sse << endl;

	} // iter

	return ssevals;
}




nn_mlp_tool::Matrix nn_mlp_tool::trainQprop(Matrix& pat, double learningRate, double momentum, double maxFactor) throw (mimasexception)
{
	double sse;
	
	// check that pat is initialised and the number of columns is equivalent
	// to the number of input nodes + output nodes
   MMERROR( network_initialised &&
            (numnodes[0]+numnodes[numlayers-1]==(signed)pat.size2()),
            mimasexception, ,
            "nn_mlp_tool::train - input/output pattern dimensions not "
            "consistent" );


	// initialise accumulators
	shared_array< Matrix > gradErrorWeight( new Matrix[numlayers] );
	shared_array< Vector > gradErrorBias( new Vector[numlayers] );
	shared_array< Matrix > prevGradErrorWeight( new Matrix [numlayers] );
   shared_array< Vector > prevGradErrorBias( new Vector[numlayers] );

	// Zero previous gradients
	for (int l=1; l<numlayers; l++)
	{
		prevGradErrorWeight[l].resize(numnodes[l],numnodes[l-1]);
		prevGradErrorBias[l].resize(numnodes[l]);
	}

	// Set some initial value for prevWeightChange, otherwise update will always be zero!
	for (int l=1; l<numlayers; l++)
	{
		gradErrorWeight[l].resize(numnodes[l],numnodes[l-1]);
		gradErrorBias[l].resize(numnodes[l]);

      prevWeightChange[l].resize( numnodes[l], numnodes[l-1] );
      fill( prevWeightChange[l].data().begin(),
            prevWeightChange[l].data().end(), 0.1 );

      prevBiasChange[l].resize( numnodes[l] );
      fill( prevBiasChange[l].data().begin(),
            prevBiasChange[l].data().end(), 0.1 );

      prevGradErrorWeight[l].resize( numnodes[l],numnodes[l-1] );
      fill( prevGradErrorWeight[l].data().begin(),
            prevGradErrorWeight[l].data().end(), 0.1 );

      prevGradErrorBias[l].resize( numnodes[l] );
      fill( prevGradErrorBias[l].data().begin(),
            prevGradErrorBias[l].data().end(), 0.1 );
	}


	Matrix ssevals(iterations,2);

	for (int iter=0; iter<iterations; iter++)
	{

		sse=0.0;
		
		// Zero current gradients
		for (int l=1; l<numlayers; l++)
		{
         fill( gradErrorWeight[l].data().begin(),
               gradErrorWeight[l].data().end(), 0.0 );
         fill( gradErrorBias[l].data().begin(),
               gradErrorBias[l].data().end(), 0.0 );
		}


		// loop through all patterns
		for (int p=0; p<(signed)pat.size1(); p++)
		{
			sse+=deltaRule(pat,p);

			// calculate gradients of error wrt weights
			for (int l=1; l<numlayers; l++)
				for (int k=0; k<numnodes[l]; k++)
				{
					for (int j=0; j<numnodes[l-1]; j++)
						gradErrorWeight[l](k,j)=gradErrorWeight[l](k,j)+(delta[l](k)*y[l-1](j));

					gradErrorBias[l](k)=gradErrorBias[l](k)+(delta[l](k)*1.0);
				}


		} // pat


		// Perform weight update
		double weightupdate; // used to speed-up computation

		// update weights in all layers
		for (int l=1; l<numlayers; l++)
		{
			for (int k=0; k<numnodes[l]; k++)
			{
				for (int j=0; j<numnodes[l-1]; j++)
				{
					weightupdate=0.0;

					// normalise over all patterns
					gradErrorWeight[l](k,j)=gradErrorWeight[l](k,j)/pat.size1();

					if (prevWeightChange[l](k,j)>0)
					{
						if (gradErrorWeight[l](k,j)>0) weightupdate=learningRate*gradErrorWeight[l](k,j);

						if (gradErrorWeight[l](k,j) > (maxFactor/(1.0+maxFactor))*prevGradErrorWeight[l](k,j) )
							weightupdate=weightupdate + maxFactor*prevWeightChange[l](k,j);
						else
							weightupdate=weightupdate + ( (gradErrorWeight[l](k,j)*prevWeightChange[l](k,j)) / (prevGradErrorWeight[l](k,j) - gradErrorWeight[l](k,j)));
						
					}
					else if (prevWeightChange[l](k,j)<0)
					{
						if (gradErrorWeight[l](k,j)<0) weightupdate=learningRate*gradErrorWeight[l](k,j);

						if (gradErrorWeight[l](k,j) < (maxFactor/(1.0+maxFactor))*prevGradErrorWeight[l](k,j) )
							weightupdate=weightupdate + maxFactor*prevWeightChange[l](k,j);
						else
							weightupdate=weightupdate + ( (gradErrorWeight[l](k,j)*prevWeightChange[l](k,j)) / (prevGradErrorWeight[l](k,j) - gradErrorWeight[l](k,j)));
					}
					else
					{
						weightupdate=learningRate*gradErrorWeight[l](k,j) + momentum*prevWeightChange[l](k,j);
					}

							
					// update weight
					weight[l](k,j)=weight[l](k,j)+weightupdate;

					// store weight updates and previous gradients
					prevWeightChange[l](k,j)=weightupdate;
					prevGradErrorWeight[l](k,j)=gradErrorWeight[l](k,j);


				}

				// normalise over all patterns
				gradErrorBias[l](k)=gradErrorBias[l](k)/pat.size1();

				weightupdate=0.0;

				if (prevBiasChange[l](k)>0)
				{
					if (gradErrorBias[l](k)>0) weightupdate=learningRate*gradErrorBias[l](k);

					if (gradErrorBias[l](k) > (maxFactor/(1.0+maxFactor))*prevGradErrorBias[l](k) )
						weightupdate=weightupdate + maxFactor*prevBiasChange[l](k);
					else
						weightupdate=weightupdate + ( (gradErrorBias[l](k)*prevBiasChange[l](k)) / (prevGradErrorBias[l](k) - gradErrorBias[l](k)));

				}
				else if (prevBiasChange[l](k)<0)
				{
					if (gradErrorBias[l](k)<0) weightupdate=learningRate*gradErrorBias[l](k);

					if (gradErrorBias[l](k) < (maxFactor/(1.0+maxFactor))*prevGradErrorBias[l](k) )
						weightupdate=weightupdate + maxFactor*prevBiasChange[l](k);
					else
						weightupdate=weightupdate + ( (gradErrorBias[l](k)*prevBiasChange[l](k)) / (prevGradErrorBias[l](k) - gradErrorBias[l](k)));
				}
				else
				{
					weightupdate=learningRate*gradErrorBias[l](k) + momentum*prevBiasChange[l](k);
				}

				// update weight
				bias[l](k)=bias[l](k)+weightupdate;

				// store weight updates and previous gradients
				prevBiasChange[l](k)=weightupdate;
				prevGradErrorBias[l](k)=gradErrorBias[l](k);


			}
		}

		ssevals(iter,0)=(double)iter;
		ssevals(iter,1)=sse;

		if (debugtrain) cout << iter << " \t" << sse << endl;

	} // iter

	return ssevals;
}



nn_mlp_tool::Matrix nn_mlp_tool::trainRprop(Matrix& pat, double etapos, double etaneg)
  throw (mimasexception)
{

	double sse;
	double delmax=50.0;
	double delmin=0.000001;
	
	// check that pat is initialised and the number of columns is equivalent
	// to the number of input nodes + output nodes
   MMERROR( network_initialised &&
            (numnodes[0]+numnodes[numlayers-1]!=(signed)pat.size2()),
            mimasexception, ,
            "nn_mlp_tool::train - input/output pattern dimensions not "
            "consistent" );


	// initialise accumulators
	shared_array< Matrix > gradErrorWeight( new Matrix[numlayers] );
	shared_array< Vector > gradErrorBias( new Vector[numlayers] );
	shared_array< Matrix > prevGradErrorWeight( new Matrix[numlayers] );
   shared_array< Vector > prevGradErrorBias( new Vector[numlayers] );

	// initialise Rprop stores
	shared_array< Matrix > prevUpdateWeight( new Matrix[numlayers] );
   shared_array< Vector > prevUpdateBias( new Vector[numlayers] );

	// Initialise and zero previous gradients
	for (int l=1; l<numlayers; l++)
	{
      gradErrorWeight[l].resize(numnodes[l],numnodes[l-1]);
		gradErrorBias[l].resize(numnodes[l]);
		prevGradErrorWeight[l].resize(numnodes[l],numnodes[l-1]);
		prevGradErrorBias[l].resize(numnodes[l]);
		prevUpdateWeight[l].resize(numnodes[l],numnodes[l-1]);
		prevUpdateBias[l].resize(numnodes[l]);
	}

	// Set some initial value for prevWeightChange, otherwise update will always be zero!
	for (int l=1; l<numlayers; l++)
	{
      fill( prevWeightChange[l].data().begin(),
            prevWeightChange[l].data().end(), 0.01 );
      fill( prevBiasChange[l].data().begin(),
            prevBiasChange[l].data().end(), 0.01 );
      fill( prevUpdateWeight[l].data().begin(),
            prevUpdateWeight[l].data().end(), 0.1 );
      fill( prevUpdateBias[l].data().begin(),
            prevUpdateBias[l].data().end(), 0.1 );
	}

	Matrix ssevals( iterations, 2 );

	for (int iter=0; iter<iterations; iter++)
	{

		sse=0.0;
		
		// Zero current gradients
		for (int l=1; l<numlayers; l++)
		{
			// don't use zeroes function here as it is slower
        fill( gradErrorWeight[l].data().begin(),
              gradErrorWeight[l].data().end(), 0.0 );
        fill( gradErrorBias[l].data().begin(),
              gradErrorBias[l].data().end(), 0.0 );
		}

		// loop through all patterns
		for (int p=0; p<(signed)pat.size1(); p++)
		{
			sse+=deltaRule(pat,p);

			// calculate gradients of error wrt weights
			for (int l=1; l<numlayers; l++)
				for (int k=0; k<numnodes[l]; k++)
				{
					for (int j=0; j<numnodes[l-1]; j++)
						gradErrorWeight[l](k,j)=gradErrorWeight[l](k,j)+(delta[l](k)*y[l-1](j));

					gradErrorBias[l](k)=gradErrorBias[l](k)+(delta[l](k)*1.0);
				}


		} // pat


		// Perform weight update
		double weightupdate; // used to speed-up computation
		// update weights in all layers
		for (int l=1; l<numlayers; l++)
		{
			for (int k=0; k<numnodes[l]; k++)
			{
				for (int j=0; j<numnodes[l-1]; j++)
				{
					// for weights in all layers

					// adapt the step size
					weightupdate=0.0;
					
					if ((gradErrorWeight[l](k,j)*prevGradErrorWeight[l](k,j))>0.0)
					{
						weightupdate=prevUpdateWeight[l](k,j)*etapos;
						if (weightupdate>delmax) weightupdate=delmax;

						weight[l](k,j)=weight[l](k,j)+((gradErrorWeight[l](k,j)>0.0?1.0:-1.0)*weightupdate);
						prevGradErrorWeight[l](k,j)=gradErrorWeight[l](k,j);
						
						
					}
					else if ((gradErrorWeight[l](k,j)*prevGradErrorWeight[l](k,j))<0.0)
					{
						weightupdate=prevUpdateWeight[l](k,j)*etaneg;
						if (weightupdate<delmin) weightupdate=delmin;
						prevGradErrorWeight[l](k,j)=0.0;
						
					}
					else
					{
						weightupdate=delmin; // this is not present in Reidmiller's paper but it is logical
						weight[l](k,j)=weight[l](k,j)+((gradErrorWeight[l](k,j)>0.0?1.0:-1.0)*weightupdate);
						prevGradErrorWeight[l](k,j)=gradErrorWeight[l](k,j);
						
					}
					prevUpdateWeight[l](k,j)=weightupdate;
						
					// weightupdate in this case is not the weight change but rather the delta parameter in Reidmiller's paper
					//prevWeightChange[l](k,j)=weightupdate;
				}


				// for biases in all layers
				weightupdate=0.0;

				if ((gradErrorBias[l](k)*prevGradErrorBias[l](k))>0.0)
				{
					weightupdate=prevUpdateBias[l](k)*etapos;
					if (weightupdate>delmax) weightupdate=delmax;

					bias[l](k)=bias[l](k)+((gradErrorBias[l](k)>0.0?1.0:-1.0)*weightupdate);
					prevGradErrorBias[l](k)=gradErrorBias[l](k);


				}
				else if ((gradErrorBias[l](k)*prevGradErrorBias[l](k))<0.0)
				{
					weightupdate=prevUpdateBias[l](k)*etaneg;
					if (weightupdate<delmin) weightupdate=delmin;
					prevGradErrorBias[l](k)=0.0;

				}
				else
				{
					weightupdate=delmin; // this is not present in Reidmiller's paper but it is logical
					bias[l](k)=bias[l](k)+((gradErrorBias[l](k)>0.0?1.0:-1.0)*weightupdate);
					prevGradErrorBias[l](k)=gradErrorBias[l](k);

				}
				prevUpdateBias[l](k)=weightupdate;
				
				//prevBiasChange[l](k,0)=weightupdate;

			}
		}

		ssevals(iter,0)=(double)iter;
		ssevals(iter,1)=sse;

		if (debugtrain) cout << iter << " \t" << sse << endl;

	} // iter

	return ssevals;
}



// set/unset debugging flag. When set, training informaiton will be displayed
void nn_mlp_tool::setDebug(bool val)
{
	debugtrain=val;
}

}
